﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Data;


namespace OlapAxisWithXMLA
{
    public static class ApplicationServices
    {

        //const string olapConnectionString = "Provider=MSOLAP.3;Persist Security Info=True;Initial Catalog=Adventure Works DW Standard Edition;Data Source=http://sampledata.infragistics.com/olap/msmdpump.dll;MDX Compatibility=1;Safety Options=2;MDX Missing Member Mode=Error";
        const string olapConnectionString = "Provider=MSOLAP.3;Persist Security Info=True;Initial Catalog=AdventureWorks;Data Source=DVFLDT136\\SQLSERVER2012;MDX Compatibility=1;Safety Options=2;MDX Missing Member Mode=Error";



        #region Private Functions

        private static uint IntValue(uint item)
        {
            return item;
        }

        private static string StringOrNull(object item)
        {
            if (item != null)
                return item.ToString();
            return null;
        }

        #endregion

        public static List<AdomdCube> LoadCubes()
        {

            var conn = OpenNewConnection();
            var cubeDefsCollection = new List<AdomdCube>();

            foreach (var cube in conn.Cubes)
            {
                if (cube.Type == Microsoft.AnalysisServices.AdomdClient.CubeType.Cube)
                {
                    cubeDefsCollection.Add(
                        new AdomdCube
                        {
                            CubeName = cube.Name,
                            Caption = cube.Caption,
                            Description = cube.Description,
                            LastDataUpdate = cube.LastUpdated,
                            CubeType = cube.Type.ToString()
                        });
                }
            }

            return cubeDefsCollection;
        }

        public static string LoadCubeSchema()
        {
            var conn = OpenNewConnection();
            return conn.Database;
        }
        public static List<AdomdDimension> LoadCubeDimensions(string cubeName)
        {
            var conn = OpenNewConnection();
            var cubeDef = conn.Cubes[cubeName];

            if (cubeDef == null)
            {
                return null;
            }

            var dimensions =
                new List<AdomdDimension>();

            foreach (var dimensionData in cubeDef.Dimensions)
            {
                if (dimensionData.DimensionType == Microsoft.AnalysisServices.AdomdClient.DimensionTypeEnum.Measure)
                {
                    continue;
                }

                var currentDimension =
                    new AdomdDimension
                    {
                        Name = dimensionData.Name,
                        UniqueName = dimensionData.UniqueName,
                        Caption = dimensionData.Caption,
                        Description = dimensionData.Description,
                        DimensionType = dimensionData.DimensionType.ToString()
                    };

                foreach (var hierarchyData in dimensionData.Hierarchies)
                {

                    var currentHierarchy =
                        new AdomdHierarchy
                        {
                            Name = hierarchyData.Name,
                            UniqueName = hierarchyData.UniqueName,
                            Caption = hierarchyData.Caption,
                            Description = hierarchyData.Description,
                            DefaultMember = hierarchyData.DefaultMember,
                            DimensionUniqueName = currentDimension.UniqueName
                        };

                    foreach (var level in hierarchyData.Levels)
                    {

                        var currentLevel =
                        new AdomdLevel
                        {
                            Name = level.Name,
                            Caption = level.Caption,
                            UniqueName = level.UniqueName,
                            Description = level.Description,
                            Number = (uint)level.LevelNumber,
                            HierarchyUniqueName = currentHierarchy.Name,
                            MembersCount = level.MemberCount
                        };

                        currentHierarchy.Levels.Add(currentLevel);
                    }

                    currentDimension.Hierarchies.Add(currentHierarchy);
                }

                dimensions.Add(currentDimension);
            }

            return dimensions;
        }

        /// <summary>
        /// Loads the cube measures.
        /// </summary>
        /// <param name="cubeName">Name of the cube.</param>
        /// <returns></returns>
        public static List<AdomdMeasure> LoadCubeMeasures(string cubeName)
        {
            var conn = OpenNewConnection();
            var cubeDef = conn.Cubes[cubeName];

            if (cubeDef == null)
            {
                return null;
            }

            var measures = new List<AdomdMeasure>();

            foreach (var measure in cubeDef.Measures)
            {
                var measureGroupName = measure.Properties["MEASUREGROUP_NAME"].Value.ToString();
                var aggregator = measure.Properties["MEASURE_AGGREGATOR"].Value.ToString();
                var currentMeasure = new AdomdMeasure
                {
                    Name = measure.Name,
                    UniqueName = measure.UniqueName,
                    Caption = measure.Caption,
                    Description = measure.Description,
                    MeasureGroupName = measureGroupName,
                    Aggregator = aggregator
                };

                measures.Add(currentMeasure);
            }

            return measures;
        }

        /// <summary>
        /// Loads the cube members.
        /// </summary>
        /// <param name="cubeName">Name of the cube.</param>
        /// <param name="hierarchyName">Name of the hierarchy.</param>
        /// <returns></returns>
        public static List<AdomdMember> LoadCubeMembers(string cubeName, string hierarchyName)
        {
            var conn = OpenNewConnection();
            var restrictions = new Microsoft.AnalysisServices.AdomdClient.AdomdRestrictionCollection
                                   {
                                       {"CUBE_NAME", cubeName},
                                       {"HIERARCHY_UNIQUE_NAME", hierarchyName}
                                   };

            var result = conn.GetSchemaDataSet("MDSCHEMA_MEMBERS", restrictions);

            var members = new List<AdomdMember>();
            foreach (DataRow dataRow in result.Tables[0].Rows)
            {
                var member = new AdomdMember
                {
                    Name = StringOrNull(dataRow["MEMBER_NAME"]),
                    UniqueName = StringOrNull(dataRow["MEMBER_UNIQUE_NAME"]),
                    Caption = StringOrNull(dataRow["MEMBER_CAPTION"]),
                    ParentUniqueName = StringOrNull(dataRow["PARENT_UNIQUE_NAME"]),
                    LevelUniqueName = StringOrNull(dataRow["LEVEL_UNIQUE_NAME"]),
                    LevelNumber = IntValue((uint)dataRow["LEVEL_NUMBER"]),
                    ChildrenCardinality = (int)IntValue((uint)dataRow["CHILDREN_CARDINALITY"])
                };

                members.Add(member);
            }

            return members;

        }

        /// <summary>
        /// Loads the cube measure groups.
        /// </summary>
        /// <param name="cubeName">Name of the cube.</param>
        /// <returns></returns>
        public static List<string> LoadCubeMeasureGroups(string cubeName)
        {
            var conn = OpenNewConnection();
            var cubeDef = conn.Cubes[cubeName];

            if (cubeDef == null)
            {
                return null;
            }

            var measureGroupsNames = new List<string>();
            foreach (var measure in cubeDef.Measures)
            {
                var measureGroupName = measure.Properties["MEASUREGROUP_NAME"].Value.ToString();
                if (!measureGroupsNames.Contains(measureGroupName))
                {
                    measureGroupsNames.Add(measureGroupName);
                }
            }

            return measureGroupsNames;

        }

        /// <summary>
        /// Loads the cube measure group dimensions.
        /// </summary>
        /// <param name="cubeName">Name of the cube.</param>
        /// <param name="catalogName">Name of the catalog.</param>
        /// <returns></returns>
        public static Dictionary<string, List<string>> LoadCubeMeasureGroupDimensions(string cubeName, string catalogName)
        {
            var conn = OpenNewConnection();
            var restrictions = new Microsoft.AnalysisServices.AdomdClient.AdomdRestrictionCollection { { "CATALOG_NAME", catalogName }, { "CUBE_NAME", cubeName } };

            var result = conn.GetSchemaDataSet("MDSCHEMA_MEASUREGROUP_DIMENSIONS", restrictions);
            var groupDimensions = new Dictionary<string, List<string>>();

            foreach (DataRow row in result.Tables[0].Rows)
            {
                var groupName = (string)row["MEASUREGROUP_NAME"];
                var dimensionUniqueName = (string)row["DIMENSION_UNIQUE_NAME"];

                List<string> dimensions;
                if (groupDimensions.TryGetValue(groupName, out dimensions))
                {
                    dimensions.Add(dimensionUniqueName);
                }
                else
                {
                    groupDimensions.Add(groupName, new List<string> { dimensionUniqueName });
                }
            }

            return groupDimensions;

        }

        /// <summary>
        /// Loads the cube level members.
        /// </summary>
        /// <param name="cubeName">Name of the cube.</param>
        /// <param name="levelUniqueName">Name of the level unique.</param>
        /// <returns></returns>
        public static List<AdomdMember> LoadCubeLevelMembers(string cubeName, string levelUniqueName)
        {
            var conn = OpenNewConnection();
            var cubeDef = conn.Cubes[cubeName];

            if (cubeDef == null)
            {
                return null;
            }

            var matches = System.Text.RegularExpressions.Regex.Matches(levelUniqueName, @"(\[.+?\])");
            if (matches.Count != 3)
            {
                throw new ArgumentException("Level unique name does not follow the expected format [Dimension].[Hierarchy].[Level]", "request");
            }

            var dimensionName = matches[0].Value.Replace("[", "").Replace("]", "");
            var dimension = cubeDef.Dimensions[dimensionName];

            if (dimension == null)
            {
                return null;
            }

            var hierarchyName = matches[1].Value.Replace("[", "").Replace("]", "");
            var hierarchy = dimension.Hierarchies[hierarchyName];

            if (hierarchy == null)
            {
                return null;
            }

            var levelName = matches[2].Value.Replace("[", "").Replace("]", "");
            Microsoft.AnalysisServices.AdomdClient.Level adomdLevel = hierarchy.Levels[levelName];

            if (adomdLevel == null)
            {
                return null;
            }

            var levelMembers = adomdLevel.GetMembers();
            var members = new List<AdomdMember>();

            foreach (var memberData in levelMembers)
            {
                var parentName = memberData.Parent == null ? string.Empty : memberData.Parent.Name;

                var member =
                    new AdomdMember
                    {
                        Name = memberData.Name,
                        UniqueName = memberData.UniqueName,
                        Caption = memberData.Caption,
                        ParentUniqueName = parentName,
                        LevelUniqueName = memberData.LevelName,
                        LevelNumber = (uint)memberData.LevelDepth,
                        ChildrenCardinality = (int)memberData.ChildCount
                    };

                members.Add(member);
            }

            return members;
        }

        /// <summary>
        /// Loads the cube filter member members.
        /// </summary>
        /// <param name="cubeName">Name of the cube.</param>
        /// <param name="parentMemberUniqueName">Name of the parent member unique.</param>
        /// <param name="levelNumber">The level number.</param>
        /// <returns></returns>
        public static List<AdomdMember> LoadCubeFilterMemberMembers(string cubeName, string parentMemberUniqueName, int levelNumber)
        {
            var conn = OpenNewConnection();
            var cubeDef = conn.Cubes[cubeName];

            if (cubeDef == null)
            {
                return null;
            }

            var matches = System.Text.RegularExpressions.Regex.Matches(parentMemberUniqueName, @"(\[.+?\])");
            if (matches.Count < 2)
            {
                throw new ArgumentException("Level unique name does not follow the expected format [Dimension].[Hierarchy].[Member]", "request");
            }

            var dimensionName = matches[0].Value.Replace("[", "").Replace("]", "");
            var dimension = cubeDef.Dimensions[dimensionName];

            if (dimension == null)
            {
                return null;
            }

            var hierarchyName = matches[1].Value.Replace("[", "").Replace("]", "");
            var hierarchy = dimension.Hierarchies[hierarchyName];

            if (hierarchy == null)
            {
                return null;
            }

            var adomdLevel = hierarchy.Levels[levelNumber];

            if (adomdLevel == null)
            {
                return null;
            }

            var levelMembers = adomdLevel.GetMembers();
            var members = new List<AdomdMember>();

            foreach (var memberData in levelMembers)
            {
                if (memberData.Parent.UniqueName != parentMemberUniqueName)
                {
                    continue;
                }

                var parentName = memberData.Parent == null ? string.Empty : memberData.Parent.Name;

                var member =
                    new AdomdMember
                    {
                        Name = memberData.Name,
                        UniqueName = memberData.UniqueName,
                        Caption = memberData.Caption,
                        ParentUniqueName = parentName,
                        LevelUniqueName = memberData.LevelName,
                        LevelNumber = (uint)memberData.LevelDepth,
                        ChildrenCardinality = (int)memberData.ChildCount
                    };

                members.Add(member);
            }

            return members;

        }

        public static GridData CubeExecuteCommand(string query)
        {
            var conn = OpenNewConnection();
            var command = new Microsoft.AnalysisServices.AdomdClient.AdomdCommand(query, conn);

            var cellSet = command.ExecuteCellSet();

            var gridData = new GridData { Axes = new List<AxisData>() };
            foreach (var axis in cellSet.Axes)
            {
                var axisData = new AxisData { Positions = new List<PositionData>() };

                foreach (var position in axis.Positions)
                {
                    var positionData = new PositionData
                    {
                        Ordinal = position.Ordinal,
                        Members = new List<MemberData>()
                    };

                    foreach (var member in position.Members)
                    {
                        var memberData = new MemberData
                        {
                            Caption = member.Caption,
                            ChildCount = (int)member.ChildCount,
                            LevelDepth = member.LevelDepth,
                            LevelName = member.LevelName,
                            Name = member.Name,
                            ParentUniqueName = member.Parent == null ? string.Empty : member.Parent.UniqueName,
                            UniqueName = member.UniqueName,
                            DrilledDown = member.DrilledDown,
                            ParentSameAsPrevious = member.ParentSameAsPrevious
                        };

                        positionData.Members.Add(memberData);
                    }

                    axisData.Positions.Add(positionData);
                }

                gridData.Axes.Add(axisData);
            }

            gridData.Cells = new List<CellData>();

            foreach (var cell in cellSet.Cells)
            {
                var cellData = new CellData
                {
                    FormattedValue = cell.FormattedValue,
                    Value = cell.Value
                };

                gridData.Cells.Add(cellData);
            }

            return gridData;
        }

        private static string Deprefix(string name)
        {
            for (var i = 0; i < name.Length; ++i)
                if (!Char.IsDigit(name, i))
                    return name.Substring(i).Replace("LPAR", "(").Replace("RPAR", ")");
            return string.Empty; // all digits :)
        }

        private static Microsoft.AnalysisServices.AdomdClient.AdomdConnection OpenNewConnection()
        {
            var conn = new Microsoft.AnalysisServices.AdomdClient.AdomdConnection(olapConnectionString);
            conn.Open();
            return conn;
        }

    }
}
